"use strict";

function _interopDefault(ex) {
  return ex && "object" == typeof ex && "default" in ex ? ex.default : ex;
}

Object.defineProperty(exports, "__esModule", {
  value: !0
});

var JavascriptModulesPlugin = _interopDefault(require("webpack/lib/javascript/JavascriptModulesPlugin")), Template$2 = _interopDefault(require("webpack/lib/Template")), propertyAccess = _interopDefault(require("webpack/lib/util/propertyAccess")), validateOptions = _interopDefault(require("schema-utils")), ModuleDependency = _interopDefault(require("webpack/lib/dependencies/ModuleDependency")), webpackSources = require("webpack-sources"), Dependency = _interopDefault(require("webpack/lib/Dependency")), ModuleFactory = _interopDefault(require("webpack/lib/ModuleFactory")), Module$2 = _interopDefault(require("webpack/lib/Module")), AsyncDependenciesBlock = _interopDefault(require("webpack/lib/AsyncDependenciesBlock")), RuntimeGlobals$2 = _interopDefault(require("webpack/lib/RuntimeGlobals"));

class ContainerExposedDependency extends ModuleDependency {
  constructor(name, request) {
    super(request), this._name = name;
  }
  get exposedName() {
    return this._name;
  }
  getResourceIdentifier() {
    return `exposed dependency ${this._name}`;
  }
}

class ContainerEntryDependency extends Dependency {
  constructor(dependencies, name) {
    super(), this.exposedDependencies = dependencies, this.optional = !0, this.loc = {
      name: name
    };
  }
}

const SOURCE_TYPES = new Set([ "javascript" ]), RUNTIME_REQUIREMENTS = new Set([ RuntimeGlobals$2.definePropertyGetters, RuntimeGlobals$2.exports, RuntimeGlobals$2.returnExportsFromRuntime ]);

class ContainerEntryModule extends Module$2 {
  constructor(dependency) {
    super("javascript/dynamic", null), this.expose = dependency.exposedDependencies;
  }
  getSourceTypes() {
    return SOURCE_TYPES;
  }
  identifier() {
    return `container entry ${JSON.stringify(this.expose.map(item => item.exposedName))}`;
  }
  readableIdentifier() {
    return "container entry";
  }
  needBuild(context, callback) {
    return callback(null, !this.buildMeta);
  }
  build(options, compilation, resolver, fs, callback) {
    this.buildMeta = {}, this.buildInfo = {
      strict: !0
    }, this.clearDependenciesAndBlocks();
    for (const dep of this.expose) {
      const block = new AsyncDependenciesBlock(void 0, dep.loc, dep.userRequest);
      block.addDependency(dep), this.addBlock(block);
    }
    callback();
  }
  codeGeneration({moduleGraph: moduleGraph, chunkGraph: chunkGraph, runtimeTemplate: runtimeTemplate}) {
    const sources = new Map, runtimeRequirements = RUNTIME_REQUIREMENTS, getters = [];
    for (const block of this.blocks) {
      const {dependencies: [dep]} = block, name = dep.exposedName, mod = moduleGraph.getModule(dep), request = dep.userRequest;
      let str;
      str = mod ? `return ${runtimeTemplate.blockPromise({
        block: block,
        message: request,
        chunkGraph: chunkGraph,
        runtimeRequirements: runtimeRequirements
      })}.then(${runtimeTemplate.basicFunction("", `return ${runtimeTemplate.moduleRaw({
        module: mod,
        chunkGraph: chunkGraph,
        request: request,
        weak: !1,
        runtimeRequirements: runtimeRequirements
      })}`)});` : runtimeTemplate.throwMissingModuleErrorBlock({
        request: dep.userRequest
      }), getters.push(`${Template$2.toNormalComment(`[${name}] => ${request}`)}"${name}": ${runtimeTemplate.basicFunction("", str)}`);
    }
    return sources.set("javascript", new webpackSources.ConcatSource(`\n${runtimeTemplate.supportsConst() ? "const" : "var"} __MODULE_MAP__ = {${getters.join(",")}};`, `\n${runtimeTemplate.supportsConst() ? "const" : "var"} __GET_MODULE__ = ${runtimeTemplate.basicFunction([ "module" ], "return typeof __MODULE_MAP__[module] ==='function' ? __MODULE_MAP__[module].apply(null) : Promise.reject(new Error('Module ' + module + ' does not exist.'))")};`, `\n\n${RuntimeGlobals$2.definePropertyGetters}(exports, {\n`, Template$2.indent([ `get: ${runtimeTemplate.basicFunction("", "return __GET_MODULE__")},`, `shared: ${runtimeTemplate.basicFunction([ "module", "getter" ], "__webpack_require__.shared[module] = getter;")}` ]), "});")), 
    {
      sources: sources,
      runtimeRequirements: runtimeRequirements
    };
  }
  size(type) {
    return 42;
  }
}

class ContainerEntryModuleFactory extends ModuleFactory {
  create({dependencies: [dependency]}, callback) {
    callback(null, {
      module: new ContainerEntryModule(dependency)
    });
  }
}

const {OriginalSource: OriginalSource, RawSource: RawSource} = require("webpack-sources"), Module = require("webpack/lib/Module"), RuntimeGlobals = require("webpack/lib/RuntimeGlobals"), Template = require("webpack/lib/Template"), makeSerializable = require("webpack/lib/util/makeSerializable"), getSourceForGlobalVariableExternal = (variableName, type) => (Array.isArray(variableName) || (variableName = [ variableName ]), 
`(function() { module.exports = ${type}${variableName.map(r => `[${JSON.stringify(r)}]`).join("")}; }());`), getSourceForCommonJsExternal = moduleAndSpecifiers => {
  if (!Array.isArray(moduleAndSpecifiers)) return `module.exports = require(${JSON.stringify(moduleAndSpecifiers)});`;
  const moduleName = moduleAndSpecifiers[0], objectLookup = moduleAndSpecifiers.slice(1).map(r => `[${JSON.stringify(r)}]`).join("");
  return `module.exports = require(${JSON.stringify(moduleName)})${objectLookup};`;
}, checkExternalVariable = (variableName, request, runtimeTemplate) => `if(typeof ${variableName} === 'undefined') { ${runtimeTemplate.throwMissingModuleErrorBlock({
  request: request
})} }\n`, getSourceForAmdOrUmdExternal = (id, optional, request, runtimeTemplate) => {
  const externalVariable = `__WEBPACK_EXTERNAL_MODULE_${Template.toIdentifier(`${id}`)}__`;
  return `${optional ? checkExternalVariable(externalVariable, Array.isArray(request) ? request.join(".") : request, runtimeTemplate) : ""}module.exports = ${externalVariable};`;
}, getSourceForDefaultCase = (moduleId, optional, request, runtimeTemplate) => {
  Array.isArray(request) || (request = [ request ]);
  optional && checkExternalVariable(request.join("."), runtimeTemplate);
  return Template.asString([ "module.exports = ", `(typeof null !== 'undefined') ? __webpack_require__.shared('${request}') || Promise.resolve(__webpack_require__(${moduleId})) : `, `Promise.reject("Missing Shared Module: null cannot be found when trying to override ${request}"); ` ]);
}, TYPES = new Set([ "javascript" ]), RUNTIME_REQUIREMENTS$1 = new Set([ RuntimeGlobals.module ]);

class SharedModule extends Module {
  constructor(request, type, userRequest) {
    super("javascript/dynamic", null), this.request = request, this.externalType = type, 
    this.userRequest = userRequest;
  }
  getSourceTypes() {
    return TYPES;
  }
  libIdent(options) {
    return this.userRequest;
  }
  chunkCondition(chunk, {chunkGraph: chunkGraph}) {
    return chunkGraph.getNumberOfEntryModules(chunk) > 0;
  }
  identifier() {
    return "external " + JSON.stringify(this.request);
  }
  readableIdentifier(requestShortener) {
    return "external " + JSON.stringify(this.request);
  }
  needBuild(context, callback) {
    return callback(null, !this.buildMeta);
  }
  build(options, compilation, resolver, fs, callback) {
    this.buildMeta = {}, this.buildInfo = {
      strict: !0
    }, callback();
  }
  getSourceString(runtimeTemplate, moduleGraph, chunkGraph) {
    const request = "object" != typeof this.request || Array.isArray(this.request) ? this.request : this.request[this.externalType];
    switch (console.log("externalType", this.externalType), this.externalType) {
     case "this":
     case "window":
     case "self":
      return getSourceForGlobalVariableExternal(request, this.externalType);

     case "global":
      return getSourceForGlobalVariableExternal(request, runtimeTemplate.outputOptions.globalObject);

     case "commonjs":
     case "commonjs2":
      return getSourceForCommonJsExternal(request);

     case "amd":
     case "amd-require":
     case "umd":
     case "umd2":
     case "system":
      return getSourceForAmdOrUmdExternal(chunkGraph.getModuleId(this), this.isOptional(moduleGraph), request, runtimeTemplate);

     default:
      return getSourceForDefaultCase(chunkGraph.getModuleId(this), this.isOptional(moduleGraph), request, runtimeTemplate);
    }
  }
  codeGeneration({runtimeTemplate: runtimeTemplate, moduleGraph: moduleGraph, chunkGraph: chunkGraph}) {
    const sourceString = this.getSourceString(runtimeTemplate, moduleGraph, chunkGraph), sources = new Map;
    return this.useSourceMap ? sources.set("javascript", new OriginalSource(sourceString, this.identifier())) : sources.set("javascript", new RawSource(sourceString)), 
    {
      sources: sources,
      runtimeRequirements: RUNTIME_REQUIREMENTS$1
    };
  }
  size(type) {
    return 42;
  }
  updateHash(hash, chunkGraph) {}
  serialize(context) {
    const {write: write} = context;
    write(this.request), write(this.externalType), write(this.userRequest), super.serialize(context);
  }
  deserialize(context) {
    const {read: read} = context;
    this.request = read(), this.externalType = read(), this.userRequest = read(), super.deserialize(context);
  }
}

makeSerializable(SharedModule, "webpack/lib/SharedModule");

class ContainerPlugin {
  static get name() {
    return ContainerPlugin.constructor.name;
  }
  constructor(options) {
    var _options$name, _options$shared, _options$library, _options$libraryTarge, _options$filename, _options$expose;
    const name = null !== (_options$name = options.name) && void 0 !== _options$name ? _options$name : "remoteEntry";
    validateOptions({
      type: "object",
      properties: {
        shared: {
          type: "object"
        },
        expose: {
          type: [ "object", "array" ]
        },
        name: {
          type: "string",
          default: name
        },
        library: {
          type: "string",
          default: name
        },
        libraryTarget: {
          type: "string",
          default: "global",
          enum: [ "var", "this", "window", "self", "global", "commonjs", "commonjs2", "amd", "amd-require", "umd", "umd2", "system" ]
        },
        filename: {
          anyOf: [ {
            type: "string"
          }, {
            instanceof: "Function"
          } ]
        }
      },
      additionalProperties: !1
    }, options, {
      name: ContainerPlugin.name
    }), this.options = {
      shared: null !== (_options$shared = options.shared) && void 0 !== _options$shared ? _options$shared : null,
      name: name,
      library: null !== (_options$library = options.library) && void 0 !== _options$library ? _options$library : name,
      libraryTarget: null !== (_options$libraryTarge = options.libraryTarget) && void 0 !== _options$libraryTarge ? _options$libraryTarge : "global",
      filename: null !== (_options$filename = options.filename) && void 0 !== _options$filename ? _options$filename : void 0,
      expose: null !== (_options$expose = options.expose) && void 0 !== _options$expose ? _options$expose : {}
    };
  }
  apply(compiler) {
    var _compiler$name;
    if (compiler.options.optimization.runtimeChunk) throw new Error("This plugin cannot integrate with RuntimeChunk plugin, please remote `optimization.runtimeChunk`.");
    compiler.options.output.jsonpFunction = `${compiler.options.output.jsonpFunction}${null !== (_compiler$name = compiler.name) && void 0 !== _compiler$name ? _compiler$name : ""}${this.options.name}`, 
    compiler.hooks.make.tapAsync(ContainerPlugin.name, (compilation, callback) => {
      let exposedMap = this.options.expose;
      if (Array.isArray(this.options.expose)) {
        exposedMap = {};
        for (const exp of this.options.expose) exposedMap[exp.replace(/(^(?:[^\w])+)/, "")] = exp;
      }
      compilation.addEntry(compilation.context, new ContainerEntryDependency(Object.entries(exposedMap).map(([name, request], idx) => {
        const dep = new ContainerExposedDependency(name, request);
        return dep.loc = {
          name: name,
          index: idx
        }, dep;
      }), this.options.name), this.options.name, callback);
    }), compiler.hooks.compilation.tap(ContainerPlugin.name, ({mainTemplate: mainTemplate}) => {
      mainTemplate.hooks.requireExtensions.tap("URLImportPlugin", source => Template$2.asString([ source, "__webpack_require__.shared = __webpack_require__" ]));
    }), compiler.hooks.thisCompilation.tap(ContainerPlugin.name, (compilation, {normalModuleFactory: normalModuleFactory}) => {
      compilation.dependencyFactories.set(ContainerEntryDependency, new ContainerEntryModuleFactory), 
      compilation.dependencyFactories.set(ContainerExposedDependency, normalModuleFactory), 
      JavascriptModulesPlugin.getCompilationHooks(compilation).render.tap(ContainerPlugin.name, (source, {chunk: chunk}) => {
        if (chunk.name === this.options.name) {
          const libName = Template$2.toIdentifier(compilation.getPath(this.options.library, {
            chunk: chunk
          }));
          switch (this.options.libraryTarget) {
           case "var":
            return new webpackSources.ConcatSource(`var ${libName} =`, source);

           case "this":
           case "window":
           case "self":
            return new webpackSources.ConcatSource(`${this.options.libraryTarget}${propertyAccess([ libName ])} =`, source);

           case "global":
            return new webpackSources.ConcatSource(`${compiler.options.output.globalObject}${propertyAccess([ libName ])} =`, source);

           case "commonjs":
           case "commonjs2":
            return new webpackSources.ConcatSource(`exports${propertyAccess([ libName ])} =`, source);

           case "amd":
           case "amd-require":
           case "umd":
           case "umd2":
           case "system":
           default:
            throw new Error(`${this.options.libraryTarget} is not a valid Library target`);
          }
        }
      }), compilation.hooks.afterChunks.tap(ContainerPlugin.name, chunks => {
        for (const chunk of chunks) chunk.name === this.options.name && (chunk.filenameTemplate = this.options.filename);
      });
    });
  }
}

const {OriginalSource: OriginalSource$1, RawSource: RawSource$1} = require("webpack-sources"), Module$1 = require("webpack/lib/Module"), RuntimeGlobals$1 = require("webpack/lib/RuntimeGlobals"), Template$1 = require("webpack/lib/Template"), getSourceForGlobalVariableExternal$1 = (variableName, type, requestScope) => {
  Array.isArray(variableName) || (variableName = [ variableName ]);
  const objectLookup = variableName.map(r => `${JSON.stringify(r)}`).join("");
  return Template$1.asString([ "(function() {", "module.exports =", `typeof ${type}["${requestScope}"] !== 'undefined' ? ${type}["${requestScope}"].get(${objectLookup}) : `, `Promise.reject('Missing Remote Runtime: ${type}["${requestScope}"] cannot be found when trying to import ${objectLookup}'); `, "}());" ]);
}, getSourceForCommonJsExternal$1 = (moduleAndSpecifiers, requestScope) => {
  if (!Array.isArray(moduleAndSpecifiers)) return `module.exports = require(${JSON.stringify(requestScope)}).get(${JSON.stringify(moduleAndSpecifiers)});`;
  const moduleName = moduleAndSpecifiers[0], objectLookup = moduleAndSpecifiers.slice(1).map(r => `[${JSON.stringify(r)}]`).join("");
  return `module.exports = require(${JSON.stringify(requestScope)}).get(${JSON.stringify(moduleName)})${objectLookup};`;
}, checkExternalVariable$1 = (variableName, request, runtimeTemplate) => `if(typeof ${variableName} === 'undefined') { ${runtimeTemplate.throwMissingModuleErrorBlock({
  request: request
})} }\n`, getSourceForDefaultCase$1 = (optional, request, runtimeTemplate, requestScope) => {
  Array.isArray(request) || (request = [ request ]);
  optional && checkExternalVariable$1(requestScope, request.join("."), runtimeTemplate);
  return Template$1.asString([ "module.exports = ", `typeof ${requestScope} !== 'undefined' ? ${requestScope}.get('${request}') : `, `Promise.reject("Missing Remote Runtime: ${requestScope} cannot be found when trying to import ${request}"); ` ]);
}, TYPES$1 = new Set([ "javascript" ]), RUNTIME_REQUIREMENTS$2 = new Set([ RuntimeGlobals$1.module ]);

class RemoteModule extends Module$1 {
  constructor(request, type, userRequest) {
    var _request$split, _request$split$shift, _request$split2;
    super("javascript/dynamic", null), this.requestScope = null == request ? void 0 : null === (_request$split = request.split("/")) || void 0 === _request$split ? void 0 : null === (_request$split$shift = _request$split.shift) || void 0 === _request$split$shift ? void 0 : _request$split$shift.call(_request$split), 
    this.request = null == request ? void 0 : null === (_request$split2 = request.split(`${this.requestScope}/`)) || void 0 === _request$split2 ? void 0 : _request$split2[1], 
    this.remoteType = type, this.userRequest = userRequest;
  }
  getSourceTypes() {
    return TYPES$1;
  }
  libIdent(options) {
    return this.userRequest;
  }
  chunkCondition(chunk, {chunkGraph: chunkGraph}) {
    return chunkGraph.getNumberOfEntryModules(chunk) > 0;
  }
  identifier() {
    return `remote ${JSON.stringify(this.request)}`;
  }
  readableIdentifier(requestShortener) {
    return `remote ${JSON.stringify(this.request)}`;
  }
  needBuild(context, callback) {
    return callback(null, !this.buildMeta);
  }
  build(options, compilation, resolver, fs, callback) {
    this.buildMeta = {}, this.buildInfo = {
      strict: !0
    }, callback();
  }
  getSourceString(runtimeTemplate, moduleGraph, chunkGraph) {
    const request = "object" != typeof this.request || Array.isArray(this.request) ? this.request : this.request[this.remoteType];
    switch (this.remoteType) {
     case "this":
     case "window":
     case "self":
      return getSourceForGlobalVariableExternal$1(request, this.remoteType, this.requestScope);

     case "global":
      return getSourceForGlobalVariableExternal$1(request, runtimeTemplate.outputOptions.globalObject, this.requestScope);

     case "commonjs":
     case "commonjs2":
      return getSourceForCommonJsExternal$1(request, this.requestScope);

     case "amd":
     case "amd-require":
     case "umd":
     case "umd2":
     case "system":
      throw new Error(`${this.remoteType} is not supported with ContainerReferencePlugin`);

     default:
      return getSourceForDefaultCase$1(this.isOptional(moduleGraph), request, runtimeTemplate, this.requestScope);
    }
  }
  codeGeneration({runtimeTemplate: runtimeTemplate, moduleGraph: moduleGraph, chunkGraph: chunkGraph}) {
    const sourceString = this.getSourceString(runtimeTemplate, moduleGraph, chunkGraph), sources = new Map;
    return this.useSourceMap ? sources.set("javascript", new OriginalSource$1(sourceString, this.identifier())) : sources.set("javascript", new RawSource$1(sourceString)), 
    {
      sources: sources,
      runtimeRequirements: RUNTIME_REQUIREMENTS$2
    };
  }
  size(type) {
    return 42;
  }
  updateHash(hash, chunkGraph) {
    hash.update(JSON.stringify(this.request));
  }
  serialize(context) {
    const {write: write} = context;
    write(this.request), write(this.remoteType), write(this.userRequest), super.serialize(context);
  }
  deserialize(context) {
    const {read: read} = context;
    this.request = read(), this.remoteType = read(), this.userRequest = read(), super.deserialize(context);
  }
}

const UNSPECIFIED_EXTERNAL_TYPE_REGEXP = /^[a-z0-9]+ /;

class RemoteModuleFactoryPlugin {
  constructor(type, remotes) {
    this.remoteType = type, this.remotes = remotes;
  }
  apply(normalModuleFactory) {
    const globalType = this.remoteType;
    normalModuleFactory.hooks.factorize.tapAsync("RemoteModuleFactoryPlugin", (data, callback) => {
      const dependency = data.dependencies[0], handleRemote = (value, type, callback) => {
        if (!1 === value) return callback();
        let externalConfig;
        if (externalConfig = !0 === value ? dependency.request : value, void 0 === type && UNSPECIFIED_EXTERNAL_TYPE_REGEXP.test(externalConfig)) {
          const idx = externalConfig.indexOf(" ");
          type = externalConfig.substr(0, idx), externalConfig = externalConfig.substr(idx + 1);
        }
        callback(null, new RemoteModule(externalConfig, type || globalType, dependency.request));
      }, handleExternals = (remotes, callback) => {
        if ("string" == typeof remotes) {
          var _data$request, _data$request$split, _data$request$split$s, _this$remotes;
          const requestScope = null == data ? void 0 : null === (_data$request = data.request) || void 0 === _data$request ? void 0 : null === (_data$request$split = _data$request.split("/")) || void 0 === _data$request$split ? void 0 : null === (_data$request$split$s = _data$request$split.shift) || void 0 === _data$request$split$s ? void 0 : _data$request$split$s.call(_data$request$split);
          if (null === (_this$remotes = this.remotes) || void 0 === _this$remotes ? void 0 : _this$remotes.includes(requestScope)) return handleRemote(dependency.request, void 0, callback);
        } else {
          if (Array.isArray(remotes)) {
            let i = 0;
            const next = () => {
              let asyncFlag;
              const handleExternalsAndCallback = (err, module) => err ? callback(err) : module ? void callback(null, module) : asyncFlag ? void (asyncFlag = !1) : next();
              do {
                if (asyncFlag = !0, i >= remotes.length) return callback();
                handleExternals(remotes[i++], handleExternalsAndCallback);
              } while (!asyncFlag);
              asyncFlag = !1;
            };
            return void next();
          }
          if ("object" == typeof remotes && Object.prototype.hasOwnProperty.call(remotes, dependency.request)) return handleRemote(remotes[dependency.request], void 0, callback);
        }
        callback();
      };
      handleExternals(this.remotes, callback);
    });
  }
}

class ContainerReferencePlugin {
  static get name() {
    return ContainerReferencePlugin.constructor.name;
  }
  constructor(options) {
    var _options$remoteType, _options$remotes, _options$override;
    this.options = {
      remoteType: null !== (_options$remoteType = options.remoteType) && void 0 !== _options$remoteType ? _options$remoteType : "global",
      remotes: null !== (_options$remotes = options.remotes) && void 0 !== _options$remotes ? _options$remotes : [],
      override: null !== (_options$override = options.override) && void 0 !== _options$override ? _options$override : {}
    };
  }
  apply(compiler) {
    const {remotes: remotes, remoteType: remoteType} = this.options;
    compiler.hooks.compile.tap(ContainerReferencePlugin.name, ({normalModuleFactory: normalModuleFactory}) => {
      new RemoteModuleFactoryPlugin(remoteType, remotes).apply(normalModuleFactory);
    });
  }
}

exports.ContainerPlugin = ContainerPlugin, exports.ContainerReferencePlugin = ContainerReferencePlugin;
